/*
 * This file is part of Names.
 * Copyright (C) 2016 by Thomas Chust <chust@web.de>. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.chust.names
import java.io.File
import java.util.Random
import java.util.Locale
import java.util.ResourceBundle
import com.typesafe.config.Config
import com.typesafe.config.ConfigFactory
import org.garret.perst.Storage
import org.garret.perst.StorageFactory
import org.garret.perst.FieldIndex
import javafx.application.Application as FXApplication
import javafx.collections.FXCollections
import javafx.collections.ObservableList
import javafx.beans.property.ReadOnlyObjectWrapper
import javafx.beans.property.adapter.JavaBeanStringPropertyBuilder
import javafx.geometry.Orientation
import javafx.stage.Stage
import javafx.stage.FileChooser
import javafx.scene.Scene
import javafx.scene.image.ImageView
import javafx.scene.layout.BorderPane
import javafx.scene.control.ToolBar
import javafx.scene.control.Button
import javafx.scene.control.TextInputDialog
import javafx.scene.control.Separator
import javafx.scene.control.TableView
import javafx.scene.control.SelectionMode
import javafx.scene.control.TableColumn
import javafx.scene.control.cell.TextFieldTableCell
import javafx.util.converter.DefaultStringConverter

/**
 * JavaFX application entry point.
 */
class Application() : FXApplication() {
  /**
   * Application configuration.
   */
  lateinit var config: Config

  /**
   * Database handle.
   */
  lateinit var store: Storage

  /**
   * Index of names.
   */
  lateinit var infosByName: FieldIndex<NameInfo>

  /**
   * Localized strings.
   */
  lateinit var messages: ResourceBundle

  /**
   * Random number generator.
   */
  val random = Random()

  /**
   * Names to display.
   */
  val namesOnDisplay: ObservableList<NameInfo> = FXCollections.observableArrayList()

  /**
   * Initialize application resources.
   */
  override fun init() {
    super.init()

    val home =
      File(System.getProperty("user.home", "."))
      .resolve(System.getProperty("names.home", ".names"))
    if (home.mkdir())
      File(home, "names.conf")
      .writeText(javaClass.getResource("/reference.conf").readText())

    config = ConfigFactory.load(
        ConfigFactory.parseFileAnySyntax(File(home, "names")))
    config.checkValid(
        ConfigFactory.defaultReference(),
        "perst", "names")

    store = StorageFactory.getInstance().createStorage()
    val storeConfig = config.withOnlyPath("perst").withoutPath("perst.page.pool.size")
    for ((key, value) in storeConfig.entrySet()) store.setProperty(key, value.unwrapped())
    store.open(
        File(home, "names.perst").absolutePath,
        config.getBytes("perst.page.pool.size"))

    store.withTransaction {
      val root = store.getRoot<FieldIndex<NameInfo>>()
      if (root == null) {
        infosByName = store.createRandomAccessFieldIndex(
            NameInfo::class.java, "name",
            /*unique =*/ true, /*caseInsensitive =*/ true)
        store.setRoot(infosByName)
      }
      else
        infosByName = root

      this.javaClass.getResourceAsStream("names.txt").reader().forEachLine {
        if (infosByName[it] == null) infosByName.put(NameInfo(store).apply {
          name = it
        })
      }
    }

    val displayLocale = config.getString("names.locale.display")
    if (displayLocale != "DEFAULT") Locale.setDefault(
        Locale.Category.DISPLAY, Locale.forLanguageTag(displayLocale))

    val formatLocale = config.getString("names.locale.format")
    if (formatLocale != "DEFAULT") Locale.setDefault(
        Locale.Category.FORMAT, Locale.forLanguageTag(formatLocale))

    messages = ResourceBundle.getBundle(
        "${javaClass.`package`.name}.messages",
        Locale.getDefault(Locale.Category.DISPLAY))
  }

  /**
   * Randomly select some names for display.
   */
  fun displayRandomNames() {
    val count = config.getInt("names.listLength")
    store.withTransaction(readOnly = true) {
      namesOnDisplay.setAll((1..count).map {
        infosByName.getAt(random.nextInt(infosByName.size))
      })
    }
  }

  /**
   * Select names similar to a given one for display.
   * @param infos Reference name infos.
   */
  fun displaySimilarNames(infos: Iterable<NameInfo>) {
    val distance = config.getInt("names.similarDistance")
    namesOnDisplay.addAll(
        store.withTransaction(readOnly = false) {
          infos.flatMap { info ->
            if (info.similar.isEmpty()) info.addSimilar(infosByName, distance)
            info.similar
          }
        })
  }

  /**
   * Select names matching a prefix string for display.
   * @param prefix Key prefix to search.
   */
  fun displayMatchingNames(prefix: String) {
    namesOnDisplay.setAll(
        store.withTransaction(readOnly = true) {
          infosByName.getPrefixList(prefix)
        })
  }

  /**
   * Import names from a text file into the database. Display the imported names.
   * @param file Source file.
   */
  fun importNames(file: File) {
    namesOnDisplay.setAll(
        store.withTransaction(readOnly = false) {
          file.useLines { lines ->
            lines.map(String::trim)
            .filter { !(it.isEmpty() || it.startsWith("#")) }
            .map {
              val fields = it.split('\t', limit = 2)
              val info = NameInfo(store)
              info.name = fields[0]
              if (fields.size > 1) info.notes = fields[1]
              if (infosByName.put(info)) info else null
            }
            .filterNotNull()
            .toList()
          }
        })
  }

  /**
   * Export names from the database into a text file.
   * @param file Target file.
   * @param infos The names to export.
   */
  fun exportNames(file: File, infos: Iterable<NameInfo> = infosByName) {
    store.withTransaction(readOnly = true) {
      file.bufferedWriter().use { wr ->
        for (info in infos) {
          wr.write(info.name)
          wr.write("\t")
          wr.write(info.notes)
          wr.newLine()
        }
      }
    }
  }

  /**
   * Setup stage for initial application screen.
   * @param stage A stage prepared for the application.
   */
  override fun start(stage: Stage) {
    displayRandomNames()

    val nameColumn = TableColumn<NameInfo, String>(messages.getString("nameColumn"))
    nameColumn.prefWidth = 78.0
    nameColumn.setCellValueFactory {
      ReadOnlyObjectWrapper(it.value.name)
    }

    val notesColumn = TableColumn<NameInfo, String>(messages.getString("notesColumn"))
    notesColumn.prefWidth = 158.0
    notesColumn.setCellValueFactory {
      JavaBeanStringPropertyBuilder.create()
      .bean(it.value)
      .name("notes")
      .build()
    }
    notesColumn.setCellFactory {
      TextFieldTableCell<NameInfo, String>(DefaultStringConverter())
    }
    val superOnEditCommit = notesColumn.onEditCommit
    notesColumn.setOnEditCommit {
      store.withTransaction { superOnEditCommit.handle(it) }
    }

    val nameTable = TableView(namesOnDisplay)
    nameTable.isEditable = true
    nameTable.selectionModel.selectionMode = SelectionMode.MULTIPLE
    nameTable.columns.addAll(
        nameColumn,
        notesColumn)

    val displayRandomNamesButton = Button(
        messages.getString("displayRandomNamesButton"),
        ImageView(javaClass.getResource("wand.png").toString()))
    displayRandomNamesButton.setOnAction {
      displayRandomNames()
    }

    val displaySimilarNamesButton = Button(
        messages.getString("displaySimilarNamesButton"),
        ImageView(javaClass.getResource("sort-plus.png").toString()))
    displaySimilarNamesButton.setOnAction {
      if (!nameTable.selectionModel.isEmpty)
        displaySimilarNames(nameTable.selectionModel.selectedItems)
    }

    val displayMatchingNamesButton = Button(
        messages.getString("displayMatchingNamesButton"),
        ImageView(javaClass.getResource("magnifier.png").toString()))
    displayMatchingNamesButton.setOnAction {
      val textInput = TextInputDialog()
      textInput.title = messages.getString("displayMatchingNamesTitle")
      textInput.headerText = messages.getString("displayMatchingNamesText")
      val prefix = textInput.showAndWait()
      prefix.ifPresent { displayMatchingNames(it) }
    }

    val importNamesButton = Button(
        messages.getString("importNamesButton"),
        ImageView(javaClass.getResource("table-import.png").toString()))
    importNamesButton.setOnAction {
      val fileChooser = FileChooser()
      fileChooser.title = messages.getString("importNamesTitle")
      fileChooser.extensionFilters.addAll(
          FileChooser.ExtensionFilter(messages.getString("textFiles"), "*.txt"),
          FileChooser.ExtensionFilter(messages.getString("allFiles"), "*.*"))
      val file = fileChooser.showOpenDialog(stage)
      if (file != null) importNames(file)
    }

    val exportNamesButton = Button(
        messages.getString("exportNamesButton"),
        ImageView(javaClass.getResource("table-export.png").toString()))
    exportNamesButton.setOnAction {
      val fileChooser = FileChooser()
      fileChooser.title = messages.getString("exportNamesTitle")
      fileChooser.extensionFilters.addAll(
          FileChooser.ExtensionFilter(messages.getString("textFiles"), "*.txt"),
          FileChooser.ExtensionFilter(messages.getString("allFiles"), "*.*"))
      val file = fileChooser.showSaveDialog(stage)
      if (file != null) {
        if (!nameTable.selectionModel.isEmpty)
          exportNames(file, nameTable.selectionModel.selectedItems)
        else
          exportNames(file)
      }
    }

    val toolBar = ToolBar(
        displayRandomNamesButton,
        displaySimilarNamesButton,
        displayMatchingNamesButton,
        Separator(Orientation.VERTICAL),
        importNamesButton,
        exportNamesButton)

    val borderPane = BorderPane()
    borderPane.top = toolBar
    borderPane.center = nameTable
    borderPane.setPrefSize(240.0, 320.0)

    stage.title = messages.getString("applicationTitle")
    stage.scene = Scene(borderPane)
    stage.sizeToScene()
    stage.show()
  }

  /**
   * Cleanup application resources.
   */
  override fun stop() {
    store.withTransaction { store.gc() }
    store.close()
    super.stop()
  }
}

/* vim:set fdm=indent sts=2 sw=2 et: */
