/*
 * This file is part of Names.
 * Copyright (C) 2016 by Thomas Chust <chust@web.de>. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.chust.names
import org.garret.perst.Storage
import org.garret.perst.Link
import org.garret.perst.Persistent

/**
 * Compute the Levenshtein edit distance between two strings.
 * @param s First string.
 * @param t Second string.
 * @return The minimal number of one-character modifications required to turn
 * [s] into [t].
 */
fun levenshtein(s: String, t: String): Int = when {
  s == t -> 0
  s.isEmpty() -> t.length
  t.isEmpty() -> s.length
  else -> {
    val v0 = IntArray(t.length + 1) { it }
    val v1 = IntArray(t.length + 1)
    val cb = IntArray(3)

    for (i in s.indices) {
      v1[0] = i + 1

      for (j in t.indices) {
        cb[0] = v1[j] + 1
        cb[1] = v0[j + 1] + 1
        cb[2] = v0[j] + if (s[i] == t[j]) 0 else 1
        v1[j + 1] = cb.min()!!
      }

      System.arraycopy(v1, 0, v0, 0, v0.size)
    }

    v1.last()
  }
}

/**
 * Run database operations in a transaction.
 * @param readOnly Whether the transaction is read-only.
 * @param thunk Operation to perform.
 * @return Whatever the operation returns.
 */
inline fun <T> Storage.withTransaction(readOnly: Boolean = false, thunk: () -> T): T {
  beginThreadTransaction(
      if (readOnly)
        Storage.READ_ONLY_TRANSACTION
      else
        Storage.READ_WRITE_TRANSACTION)
  try {
    val result = thunk()
    endThreadTransaction()
    return result
  }
  catch (exn: Exception) {
    rollbackThreadTransaction()
    throw exn
  }
}

/**
 * Database entity for a name.
 * @param storage Database where the object lives.
 */
class NameInfo(storage: Storage) : Persistent(storage) {
  /**
   * The name itself.
   */
  var name: String = ""
    set(value) {
      field = value
      modify()
    }

  /**
   * Links to names considered similar.
   */
  val similar: Link<NameInfo> = storage.createLink()

  /**
   * Some descriptive text.
   */
  var notes: String = ""
    set(value) {
      field = value
      modify()
    }

  /**
   * Add names with a small edit distance to the list of similar names.
   * @param candidates Names to examine.
   * @param distance Maximal edit distance for similar names.
   */
  fun addSimilar(candidates: Iterable<NameInfo>, distance: Int) {
    if (distance > 0) {
      similar += candidates.filter { that ->
        val d = levenshtein(this.name, that.name)
        0 < d && d <= distance
      }
      modify()
    }
  }
}

/* vim:set fdm=indent sts=2 sw=2 et: */
